#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# run execute.py FRA Ma 3.5 1 True 8 0 1 21

"""
(c) 2020 Miguel Colom
http://mcolom.info

Under GNU GPL license. See LICENCE file for details.
"""

import sys
import os
import urllib.request
import datetime
import time
import argparse
from os.path import join, isfile
from shutil import copyfile
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from numpy import array, sqrt, median, std, insert, add, arange, ones, roll
import pandas as pd
from scipy.stats import linregress
from dataclasses import dataclass

# Constants of the methods
@dataclass
class Constants:
    a: float = 1
    b: float = 5
    tau: int = 7
    w: int = 7
    lambs: object = None

# Constants of this script
@dataclass
class ScriptConstants:
    NWeeks: int = 8 # number of weeks to compute the weekly bias correction factors
    
    alpha = 1
    alpha_grid = 0.3
    
    plot_confidence_Rt = True
    plot_confidence_Cori = False
    
    
    # Plot on the left
    plot_color_Rt = np.array(((255, 0, 0))) / 255
    plot_color_EpiEstim = np.array((70, 70, 70)) / 255

    # Plot on the right
    plot_color_i = np.array((97, 194, 0)) / 255
    plot_color_if = np.array(((255, 0, 0))) / 255
    plot_color_wnf = np.array((27, 50, 180)) / 255

    # Confidence interval colors
    conf_color_Rt = {\
'CI 95': np.array(((255,0,0))) / 255,
'CI 90': np.array(((0,255,0))) / 255 \
}
    confidence_alpha = 0.15

    
    plot_label_i = "incidence"
    plot_label_wnf = "filtered inc."
    plot_label_if = "reconstructed inc."

def code_to_country_name(code):
    d = {'agg_EU': 'Aggregated: European Union', \
'agg_south_east_asia': 'Aggregated: South East Asia', \
'AFG': 'Afghanistan', \
'ALB': 'Albania', \
'DZA': 'Algeria', \
'AND': 'Andorra', \
'AGO': 'Angola', \
'ATG': 'Antigua and Barbuda', \
'ARG': 'Argentina', \
'ARM': 'Armenia', \
'AUS': 'Australia', \
'AUT': 'Austria', \
'AZE': 'Azerbaijan', \
'BHS': 'Bahamas', \
'BHR': 'Bahrain', \
'BGD': 'Bangladesh', \
'BRB': 'Barbados', \
'BLR': 'Belarus', \
'BEL': 'Belgium', \
'BLZ': 'Belize', \
'BEN': 'Benin', \
'BTN': 'Bhutan', \
'BOL': 'Bolivia', \
'BIH': 'Bosnia and Herzegovina', \
'BWA': 'Botswana', \
'BRA': 'Brazil', \
'BRN': 'Brunei', \
'BGR': 'Bulgaria', \
'BFA': 'Burkina Faso', \
'BDI': 'Burundi', \
'KHM': 'Cambodia', \
'CMR': 'Cameroon', \
'CAN': 'Canada', \
'CPV': 'Cape Verde', \
'CAF': 'Central African Republic', \
'TCD': 'Chad', \
'CHL': 'Chile', \
'CHN': 'China', \
'COL': 'Colombia', \
'COM': 'Comoros', \
'COG': 'Congo', \
'CRI': 'Costa Rica', \
'CIV': 'Cote d\'Ivoire', \
'HRV': 'Croatia', \
'CUB': 'Cuba', \
'CYP': 'Cyprus', \
'CZE': 'Czechia', \
'COD': 'Democratic Republic of Congo', \
'DNK': 'Denmark', \
'DJI': 'Djibouti', \
'DMA': 'Dominica', \
'DOM': 'Dominican Republic', \
'ECU': 'Ecuador', \
'EGY': 'Egypt', \
'SLV': 'El Salvador', \
'GNQ': 'Equatorial Guinea', \
'ERI': 'Eritrea', \
'EST': 'Estonia', \
'SWZ': 'Eswatini', \
'ETH': 'Ethiopia', \
'FJI': 'Fiji', \
'FIN': 'Finland', \
'FRA': 'France', \
'GAB': 'Gabon', \
'GMB': 'Gambia', \
'GEO': 'Georgia', \
'DEU': 'Germany', \
'GHA': 'Ghana', \
'GRC': 'Greece', \
'GRD': 'Grenada', \
'GTM': 'Guatemala', \
'GIN': 'Guinea', \
'GNB': 'Guinea-Bissau', \
'GUY': 'Guyana', \
'HTI': 'Haiti', \
'HND': 'Honduras', \
'HKG': 'Hong Kong', \
'HUN': 'Hungary', \
'ISL': 'Iceland', \
'IND': 'India', \
'IDN': 'Indonesia', \
'IRN': 'Iran', \
'IRQ': 'Iraq', \
'IRL': 'Ireland', \
'ISR': 'Israel', \
'ITA': 'Italy', \
'JAM': 'Jamaica', \
'JPN': 'Japan', \
'JOR': 'Jordan', \
'KAZ': 'Kazakhstan', \
'KEN': 'Kenya', \
'OWID_KOS': 'Kosovo', \
'KWT': 'Kuwait', \
'KGZ': 'Kyrgyzstan', \
'LAO': 'Laos', \
'LVA': 'Latvia', \
'LBN': 'Lebanon', \
'LSO': 'Lesotho', \
'LBR': 'Liberia', \
'LBY': 'Libya', \
'LIE': 'Liechtenstein', \
'LTU': 'Lithuania', \
'LUX': 'Luxembourg', \
'MDG': 'Madagascar', \
'MWI': 'Malawi', \
'MYS': 'Malaysia', \
'MDV': 'Maldives', \
'MLI': 'Mali', \
'MLT': 'Malta', \
'MHL': 'Marshall Islands', \
'MRT': 'Mauritania', \
'MUS': 'Mauritius', \
'MEX': 'Mexico', \
'MDA': 'Moldova', \
'MCO': 'Monaco', \
'MNG': 'Mongolia', \
'MNE': 'Montenegro', \
'MAR': 'Morocco', \
'MOZ': 'Mozambique', \
'MMR': 'Myanmar', \
'NAM': 'Namibia', \
'NPL': 'Nepal', \
'NLD': 'Netherlands', \
'NZL': 'New Zealand', \
'NIC': 'Nicaragua', \
'NER': 'Niger', \
'NGA': 'Nigeria', \
'MKD': 'North Macedonia', \
'NOR': 'Norway', \
'OMN': 'Oman', \
'PAK': 'Pakistan', \
'PSE': 'Palestine', \
'PAN': 'Panama', \
'PNG': 'Papua New Guinea', \
'PRY': 'Paraguay', \
'PER': 'Peru', \
'PHL': 'Philippines', \
'POL': 'Poland', \
'PRT': 'Portugal', \
'QAT': 'Qatar', \
'ROU': 'Romania', \
'RUS': 'Russia', \
'RWA': 'Rwanda', \
'KNA': 'Saint Kitts and Nevis', \
'LCA': 'Saint Lucia', \
'VCT': 'Saint Vincent and the Grenadines', \
'WSM': 'Samoa', \
'SMR': 'San Marino', \
'STP': 'Sao Tome and Principe', \
'SAU': 'Saudi Arabia', \
'SEN': 'Senegal', \
'SRB': 'Serbia', \
'SYC': 'Seychelles', \
'SLE': 'Sierra Leone', \
'SGP': 'Singapore', \
'SVK': 'Slovakia', \
'SVN': 'Slovenia', \
'SLB': 'Solomon Islands', \
'SOM': 'Somalia', \
'ZAF': 'South Africa', \
'KOR': 'South Korea', \
'SSD': 'South Sudan', \
'ESP': 'Spain', \
'LKA': 'Sri Lanka', \
'SDN': 'Sudan', \
'SUR': 'Suriname', \
'SWE': 'Sweden', \
'CHE': 'Switzerland', \
'SYR': 'Syria', \
'TWN': 'Taiwan', \
'TJK': 'Tajikistan', \
'TZA': 'Tanzania', \
'THA': 'Thailand', \
'TLS': 'Timor', \
'TGO': 'Togo', \
'TTO': 'Trinidad and Tobago', \
'TUN': 'Tunisia', \
'TUR': 'Turkey', \
'UGA': 'Uganda', \
'UKR': 'Ukraine', \
'ARE': 'United Arab Emirates', \
'GBR': 'United Kingdom', \
'USA': 'United States', \
'URY': 'Uruguay', \
'UZB': 'Uzbekistan', \
'VUT': 'Vanuatu', \
'VAT': 'Vatican', \
'VEN': 'Venezuela', \
'VNM': 'Vietnam', \
'OWID_WRL': 'World', \
'YEM': 'Yemen', \
'ZMB': 'Zambia', \
'ZWE': 'Zimbabwe', \
'Alabama': 'USA: Alabama', \
'Alaska': 'USA: Alaska', \
'Arizona': 'USA: Arizona', \
'Arkansas': 'USA: Arkansas', \
'California': 'USA: California', \
'Colorado': 'USA: Colorado', \
'Connecticut': 'USA: Connecticut', \
'Delaware': 'USA: Delaware', \
'District_of_Columbia': 'USA: District of Columbia', \
'Florida': 'USA: Florida', \
'Georgia': 'USA: Georgia', \
'Guam': 'USA: Guam', \
'Hawaii': 'USA: Hawaii', \
'Idaho': 'USA: Idaho', \
'Illinois': 'USA: Illinois', \
'Indiana': 'USA: Indiana', \
'Iowa': 'USA: Iowa', \
'Kansas': 'USA: Kansas', \
'Kentucky': 'USA: Kentucky', \
'Louisiana': 'USA: Louisiana', \
'Maine': 'USA: Maine', \
'Maryland': 'USA: Maryland', \
'Massachusetts': 'USA: Massachusetts', \
'Michigan': 'USA: Michigan', \
'Minnesota': 'USA: Minnesota', \
'Mississippi': 'USA: Mississippi', \
'Missouri': 'USA: Missouri', \
'Montana': 'USA: Montana', \
'Nebraska': 'USA: Nebraska', \
'Nevada': 'USA: Nevada', \
'New_Hampshire': 'USA: New Hampshire', \
'New_Jersey': 'USA: New Jersey', \
'New_Mexico': 'USA: New Mexico', \
'New_York': 'USA: New York', \
'North_Carolina': 'USA: North Carolina', \
'North_Dakota': 'USA: North Dakota', \
'Northern_Mariana_Islands': 'USA: Northern Mariana Islands', \
'Ohio': 'USA: Ohio', \
'Oklahoma': 'USA: Oklahoma', \
'Oregon': 'USA: Oregon', \
'Pennsylvania': 'USA: Pennsylvania', \
'Puerto_Rico': 'USA: Puerto Rico', \
'Rhode_Island': 'USA: Rhode Island', \
'South_Carolina': 'USA: South Carolina', \
'South_Dakota': 'USA: South Dakota', \
'Tennessee': 'USA: Tennessee', \
'Texas': 'USA: Texas', \
'Utah': 'USA: Utah', \
'Vermont': 'USA: Vermont', \
'Virgin_Islands': 'USA: Virgin Islands', \
'Virginia': 'USA: Virginia', \
'Washington': 'USA: Washington', \
'West_Virginia': 'USA: West Virginia', \
'Wisconsin': 'USA: Wisconsin', \
'Wyoming': 'USA: Wyoming'}

    return d[code] if code in d else code

def get_country_name(country):
    '''
    Get the country's name from its code
    '''
    return code_to_country_name(country) if country else "uploaded data file"

def save_country(cases, deaths, start_date, end_date, population, country, data_directory):
    '''
    Save the cases of one particular country
    '''
    filename = join(data_directory, f"cases_{country}.txt")
    with open(filename, "wt") as f:
        for num_cases in cases:
            f.write(f"{num_cases}\n")

    filename = join(data_directory, f"deaths_{country}.txt")
    with open(filename, "wt") as f:
        for num_deaths in deaths:
            f.write(f"{num_deaths}\n")

    filename = join(data_directory, f"start_date_{country}.txt")
    with open(filename, "wt") as f:
        f.write(f"{start_date.timestamp()}\n")

    filename = join(data_directory, f"population_{country}.txt")
    with open(filename, "wt") as f:
        f.write(f"{population}\n")

def date_range_num(start_date, num_dates):
    '''
    Return a range of dates, from start_date with cardinal num_dates
    '''
    dates = []
    date = start_date

    for _ in range(num_dates):
        dates.append(date)
        date += datetime.timedelta(days=1)
    return dates

def date_range(start_date, end_date):
    '''
    Return a range of dates, from start_date to end_date, both included
    '''
    dates = []
    date = start_date
    while date <= end_date:
        dates.append(date)
        date += datetime.timedelta(days=1)
    return dates


def demo_failure(message):
    '''
    Exit with a demo_failure message
    '''
    print(message)
    with open("demo_failure.txt", "wt") as f:
        f.write(message)
        f.write("\n")
    sys.exit(0)

def read_daily(filename):
    '''
    Read the file with the cases
    '''
    try:
        with open(filename, "rt") as fin:
            lines = fin.readlines()

        cases = []
        for line in lines:
            if line:
                cases.append(float(line))
    except (ValueError, UnicodeDecodeError):
        demo_failure("Uploaded data: bad format.")

    return array(cases)

def read_lambdas_acc(filename):
    '''
    Read the file with the lambdas.s_Cori
    All negative days are accumulated at day 0
    '''
    acc = 0.0
    cases = []
    

    with open(filename, "rt") as f:
        lines = f.readlines()
    
    
    try:
        for line in lines:
            if line:
                day_str, value_str = line.split()
                day, value = int(day_str), float(value_str)
                
                if day < 0:
                    acc += value
                else:                
                    cases.append(value)
    except (ValueError, UnicodeDecodeError):
        demo_failure("Uploaded data: bad format.")
    
    cases[0] += acc
    return array(cases)


def read_float(filename):
    '''
    Read the file with the cases
    '''
    with open(filename, "rt") as fin:
        return float(fin.readline())

def read_country(country, data_directory):
    '''
    Read a country: deaths, infections, starting date, ...
    '''
    # Deaths
    filename = join(data_directory, f"deaths_{country}.txt")
    d = read_daily(filename)

    # Infections
    filename = join(data_directory, f"cases_{country}.txt")
    i = read_daily(filename)

    # Population
    filename = join(data_directory, f"population_{country}.txt")
    P = int(read_float(filename))

    # Dates
    filename = join(data_directory, f"start_date_{country}.txt")
    start_timestamp = read_float(filename)

    start_date = datetime.datetime.fromtimestamp(start_timestamp)
    dates = date_range_num(start_date, len(i))

    assert len(dates) == len(d) == len(i)
    return d, i, P, dates

def save_infections_dates_plot(filename, i, i_f, weekend_noise_free, dates, last_data_date, country):
    '''
    Infections plot, dates in the X axis
    '''
    plt.close("all")
    fig, ax = plt.subplots()
    plt.gcf().subplots_adjust(0.15)

    num_dates = len(dates)
    assert num_dates == len(i)

    # Skip the first points in the plot
    q = min(ScriptConstants.NWeeks*7, num_dates)
    skip = len(i) - q
    #
    i = i[skip:]
    i_f = i_f[skip:]
    if weekend_noise_free is not None:
        weekend_noise_free = weekend_noise_free[skip:]
    dates = dates[skip:]

    fig.autofmt_xdate()

    ax.plot_date(dates, i, ls='-', marker='.', \
                 label=ScriptConstants.plot_label_i, color=ScriptConstants.plot_color_i , alpha=ScriptConstants.alpha)

    if weekend_noise_free is not None:
        ax.plot_date(dates[0:len(weekend_noise_free)], weekend_noise_free, \
                     ls='--', marker='', linewidth=2, label=ScriptConstants.plot_label_wnf, \
                     color=ScriptConstants.plot_color_wnf, alpha=ScriptConstants.alpha)

    ax.plot_date(dates[0:len(i_f)], i_f, ls='-', marker='', linewidth=2, \
                 label=ScriptConstants.plot_label_if, color=ScriptConstants.plot_color_if)


    if weekend_noise_free is None:
        y_lim = 1.2*np.nanmax((i, i_f))
    else:
        y_lim = 1.2*np.nanmax((i, i_f, weekend_noise_free))
    ax.set_ylim((0, y_lim))

    ax.tick_params(axis='y', labelcolor='black')
    
    # Separate data from forecast
    if last_data_date != dates[-1]:
        ax.axvline(x=last_data_date, ls="--", color="black")
        last_data_date_str = last_data_date.strftime("%Y-%m-%d")
        ax.annotate(last_data_date_str, xy=(last_data_date, 0), \
                                         xytext=(3, 220), \
                                         textcoords='offset points')

    country_name = get_country_name(country)
    
    ax.set_title(f"Infections in {country_name}")
    ax.set_ylabel("Counts")
    ax.legend(loc="upper left", ncol=3)
    ax.grid(alpha=ScriptConstants.alpha_grid)
    plt.savefig(filename)

def save_infections_days_plot(filename, i, i_f, weekend_noise_free, last_data_index, country):
    '''
    Infections plot, day indices in the X axis
    '''    
    plt.close("all")
    fig, ax = plt.subplots()
    plt.gcf().subplots_adjust(0.15)

    len_i_orig = len(i)
    
    # Skip the first points in the plot
    q = min(ScriptConstants.NWeeks*7, len(i))
    skip = len(i) - q
    #
    i = i[skip:]
    i_f = i_f[skip:]
    if weekend_noise_free is not None:
        weekend_noise_free = weekend_noise_free[skip:]
    
    x_range = np.arange(len_i_orig-q, len_i_orig)
    
    ax.plot(x_range, i, ls='-', marker='.', label=ScriptConstants.plot_label_i, \
            color=ScriptConstants.plot_color_i, alpha=ScriptConstants.alpha)

    if weekend_noise_free is not None:
        ax.plot(x_range, weekend_noise_free, ls='--', marker='', \
                linewidth=2, label=ScriptConstants.plot_label_wnf, \
                color=ScriptConstants.plot_color_wnf, alpha=ScriptConstants.alpha)

    ax.plot(x_range, i_f, ls='-', marker='', linewidth=2, label=ScriptConstants.plot_label_if, \
            color=ScriptConstants.plot_color_if)


    if weekend_noise_free is None:
        y_lim = 1.2*np.nanmax((i, i_f))
    else:
        y_lim = 1.2*np.nanmax((i, i_f, weekend_noise_free))
    ax.set_ylim((0, y_lim))

    ax.tick_params(axis='y', labelcolor='black')
    
    # Separate data from forecast
    if len_i_orig - last_data_index != 1:
        ax.axvline(x=last_data_index, ls="--", color="black")
        ax.annotate(last_data_index, xy=(last_data_index, 0), \
                                         xytext=(3, 220), \
                                         textcoords='offset points')

    country_name = get_country_name(country)    

    ax.set_title(f"Infections in {country_name}")
    ax.set_ylabel("Counts")
    ax.legend(loc="upper left", ncol=3)
    ax.grid(alpha=ScriptConstants.alpha_grid)
    plt.savefig(filename)

def save_R_dates_plot(Rn, variability, Rn_Cori, variability_Cori, \
                      s_Cori, dates, last_data_date, country, \
                      filename="plot.png"):
    '''
    Save the Rn plot of the last days
    '''
    dates = list(dates)
    start_date = min(dates)
    end_date = max(dates)

    # Show only the last days
    q = min(ScriptConstants.NWeeks*7, len(Rn))
    Rn = Rn[-q:]
    Rn_Cori = Rn_Cori[-q:]
    start_date = end_date - datetime.timedelta(days=q-1)

    #err_Rn = get_conf_interval(Rn)
    #err_Rn = variability[-q:]
    err_Rn_Cori = variability_Cori[-q:]
    
    # We want to move Cori's curve to the left
    # A simple way to do this is to add NaNs
    for _ in range(s_Cori):
        Rn_Cori = np.append(Rn_Cori, np.nan)
        err_Rn_Cori = np.append(err_Rn_Cori, 0)
        Rn_Cori = np.delete(Rn_Cori, 0)
        err_Rn_Cori = np.delete(err_Rn_Cori, 0)

    # Do save the plot
    fig, ax = plt.subplots()
    plt.gcf().subplots_adjust(0.15)

    # Dates available: date plot
    num_dates = len(Rn)

    fig.autofmt_xdate()
    dates = date_range_num(start_date, num_dates)
    
    ax.plot_date(dates[0:len(Rn)], Rn, ls='-', marker='.', label="$R_t$", color=ScriptConstants.plot_color_Rt)
    ax.plot_date(dates[0:len(Rn_Cori)], Rn_Cori, ls='-', marker='.', label="$R_t$ (EpiEstim)", color=ScriptConstants.plot_color_EpiEstim)

    # Confidence intervals
    if ScriptConstants.plot_confidence_Rt:
        for ci_str in variability.keys():
            err_Rn = variabilities[ci_str][-q:]
            
            if len(Rn) == len(err_Rn):
                ax.fill_between(dates[0:len(Rn)], Rn-err_Rn, Rn+err_Rn, \
                                color=ScriptConstants.conf_color_Rt[ci_str], alpha=ScriptConstants.confidence_alpha)
                

    if ScriptConstants.plot_confidence_Cori:
        if len(Rn_Cori) == len(err_Rn_Cori):
            ax.fill_between(dates[0:len(Rn_Cori)], Rn_Cori-err_Rn_Cori, Rn_Cori+err_Rn_Cori, \
                            color=ScriptConstants.plot_color_EpiEstim, alpha=ScriptConstants.confidence_alpha)
        
    # Separate data from forecast
    if last_data_date != end_date:
        ax.axvline(x=last_data_date, ls="--", color="black")
        last_data_date_str = last_data_date.strftime("%Y-%m-%d")
        ax.annotate(last_data_date_str, xy=(last_data_date, 0), \
                                        xytext=(3, 220), \
                                        textcoords='offset points')
    
    # Vertical line at the end of EpiEstim
    if dates:
        ax.axvline(x=dates[len(Rn_Cori) - np.sum(np.isnan(Rn_Cori)) - 1], ls="-.", color="black", alpha=0.4)

    ax.axhline(y=1, ls="--", color="black")
    
    # Set y-axis limit depending on the data
    y_lim = np.max((2, 1.05*np.nanmax((Rn, Rn_Cori))))
    ax.set_ylim((0, y_lim))

    country_name = get_country_name(country)
    
    ax.set_title(f"Reproduction number in {country_name}")
    ax.set_ylabel("$R_t$")
    ax.legend(loc="upper left")
    ax.grid(alpha=ScriptConstants.alpha_grid)
    plt.savefig(filename)
    plt.savefig(filename.replace(".png", ".pdf"))

def save_R_days_plot(Rn, variability, Rn_Cori, variability_Cori, \
                     s_Cori, last_data_index, country, \
                     filename="plot.png"):
    '''
    Save the Rn plot of the last days
    '''    
    len_Rn_orig = len(Rn)
    len_Rn_Cori_orig = len(Rn_Cori)

    # Show only the last days
    q = min(ScriptConstants.NWeeks*7, len(Rn))
    Rn = Rn[-q:]
    Rn_Cori = Rn_Cori[-q:]

    err_Rn_Cori = variability_Cori[-q:]
    
    # We want to move Cori's curve to the left
    # A simple way to do this is to add NaNs
    for _ in range(s_Cori):
        Rn_Cori = np.append(Rn_Cori, np.nan)
        err_Rn_Cori = np.append(err_Rn_Cori, 0)
        Rn_Cori = np.delete(Rn_Cori, 0)
        err_Rn_Cori = np.delete(err_Rn_Cori, 0)

    # Do save the plot
    fig, ax = plt.subplots()
    plt.gcf().subplots_adjust(0.15)
    
    x_range = np.arange(len_Rn_orig-q, len_Rn_orig)
    #
    x_range_Cori = np.arange(len_Rn_Cori_orig-q, len_Rn_Cori_orig)
    x_range_Cori += np.max(x_range) - np.max(x_range_Cori)
    
    ax.plot(x_range, Rn, ls='-', marker='.', label="$R_t$", color=ScriptConstants.plot_color_Rt)
    ax.plot(x_range_Cori, Rn_Cori, ls='-', marker='.', label="$R_t$ (EpiEstim)", color=ScriptConstants.plot_color_EpiEstim)

    # Confidence intervals
    if ScriptConstants.plot_confidence_Rt:

        for ci_str in variability.keys():
            err_Rn = variabilities[ci_str][-q:]

            if len(Rn) == len(err_Rn):
                ax.fill_between(x_range, Rn-err_Rn, Rn+err_Rn, \
                                color=ScriptConstants.conf_color_Rt[ci_str], alpha=ScriptConstants.confidence_alpha)
    if ScriptConstants.plot_confidence_Cori:
        if len(Rn_Cori) == len(err_Rn_Cori):
            ax.fill_between(x_range_Cori, Rn_Cori-err_Rn_Cori, Rn_Cori+err_Rn_Cori, color=ScriptConstants.plot_color_EpiEstim, alpha=ScriptConstants.confidence_alpha)
        
    # Separate data from forecast
    if len_Rn_orig - last_data_index != 1:
        ax.axvline(x=last_data_index, ls="--", color="black")
        ax.annotate(last_data_index, xy=(last_data_index, 0), \
                                         xytext=(3, 220), \
                                         textcoords='offset points')

    # Vertical line at the end of EpiEstim
    if dates:
        ax.axvline(x=dates[len(Rn_Cori) - np.sum(np.isnan(Rn_Cori)) - 1],
                   ls="-.", color="black", alpha=0.4)

    ax.axhline(y=1, ls="--", color="black")
    
    # Set y-axis limit depending on the data
    y_lim = np.max((2, 1.05*np.nanmax((Rn, Rn_Cori))))
    ax.set_ylim((0, y_lim))

    country_name = get_country_name(country)

    ax.set_title(f"Reproduction number in {country_name}")
    ax.set_ylabel("$R_t$")
    ax.legend(loc="upper left")
    ax.grid(alpha=ScriptConstants.alpha_grid)
    plt.savefig(filename)
    plt.savefig(filename.replace(".png", ".pdf"))


def get_conf_interval(i_f, w=20):
    '''
    RMSE between the last w points regression and the data.
    '''

    errors = []
    for k in range(w-1):
        errors.append(0)

    for k in range(w-1, len(i_f)):
        data = i_f[k-w+1 : k+1]
        res = linregress(range(w), data)

        regressed = [res.slope * q + res.intercept for q in range(w)]

        RMSE = 2 * std(data - regressed)
        errors.append(RMSE)

    return array(errors)

def aggregate_countries(agg_countries):
    '''
    Obtain aggreagated deaths, infections, and population of a group of
    countries
    '''
    agg_dates = []
    agg_population = 0

    # Read every country
    d, i, P, dates = {}, {}, {}, {}
    for c in agg_countries:
        c_d, c_i, c_P, c_dates = read_country(c, data_directory)
        d[c] = c_d
        i[c] = c_i
        P[c] = c_P
        dates[c] = c_dates
        
    # Look for start and end dates for all countries
    agg_start_date = max([min(dates[c]) for c in agg_countries])
    agg_end_date = min([max(dates[c]) for c in agg_countries])

    agg_dates = date_range(agg_start_date, agg_end_date)
    agg_deaths = np.zeros(len(agg_dates))
    agg_cases = np.zeros(len(agg_dates))
    agg_cases_country = {}

    
    for idx_agg, date in enumerate(agg_dates):
        for c in agg_countries:
            if c not in agg_cases_country:
                agg_cases_country[c] = np.zeros(len(agg_dates))

            if date in dates[c]:
                idx_c = dates[c].index(date)
                agg_deaths[idx_agg] += d[c][idx_c]
                agg_cases[idx_agg] += i[c][idx_c]
                agg_cases_country[c][idx_agg] += i[c][idx_c]

    for c in agg_countries:
        agg_population += P[c]

    return agg_deaths, agg_cases, agg_cases_country, agg_population, agg_dates

def pandas_to_np_pyplot(df, col):
    '''
    Pandas to numpy array.
    Set nan if no value
    '''
    A = df[col].to_numpy().astype(str)
    stripped = [s.strip() for s in A]
    res = [(float(sv) if sv else np.nan) for sv in stripped]
    return array(res)


def read_program_output(filename):
    '''
    Read C++ program's output
    '''
    df = pd.read_csv(filename, sep=';')
    R = pandas_to_np_pyplot(df, 'Rt EpiInvert')

    # Tip: use df.columns to check the exact names and encoding of the columns
    #original = pandas_to_np_pyplot(df, 'original incidence')
    #filtered = pandas_to_np_pyplot(df, 'Nishiura formula')
    #variability = pandas_to_np_pyplot(df, 'variability last 3 days')
    
    original = pandas_to_np_pyplot(df, 'original incidence')
    filtered = pandas_to_np_pyplot(df, 'renewal equation')
    #variability = pandas_to_np_pyplot(df, 'Rt variability')
    
    variabilities = {}
    variabilities['CI 95'] = pandas_to_np_pyplot(df, 'CI 95')
    variabilities['CI 90'] = pandas_to_np_pyplot(df, 'CI 90')
    
    weekend_noise_free = pandas_to_np_pyplot(df, 'weekly bias correction') if 'weekly bias correction' in df.columns else None

    return R, original, filtered, variabilities, weekend_noise_free

def generate_multicolumn_aggregated(filename, agg_cases_country, agg_countries):
    one_country = list(agg_cases_country.keys())[0]
    num_cases = len(agg_cases_country[one_country])
    
    # Assert all have the same size
    for c in agg_countries:
        i = agg_cases_country[c]
        assert len(i) == num_cases
    
    with open(filename, "wt") as f:
        for i in range(num_cases):
            cases_i_by_country = np.array([agg_cases_country[c][i] for c in agg_countries]).astype(int)        
            line = " ".join(cases_i_by_country.astype(str))
            f.write(line + "\n")

def update_in_background(data_directory, background=True):
    program_path = "./background_updater.py" if isfile("./background_updater.py") else "background_updater.py"
    command = f'{program_path} --data_directory {data_directory}'
    if background:
        command += " &"
    print(f"Running: {command}")
    os.system(command)

def count_without_trailing_NaNs(A):
    '''
    Count number of elements in the array after discarding training NaNs.
    '''
    return np.where(~np.isnan(A))[0][-1] + 1

def get_max_days(Rn, i, i_f):
    '''
    Get the maximum length of the given signal, without NaNs
    '''
    r = np.max((\
      count_without_trailing_NaNs(Rn),
      count_without_trailing_NaNs(i),
      count_without_trailing_NaNs(i_f),
    ))
    return int(r)

### Cori:

def cori_numerator(t, v):
    """
    Compute Cori's method numerator
    """    
    # First index s = 0: t-tau+1 = 0 ==> t = tau - 1
    return Constants.a + sum([(v[s]) for s in range(t-Constants.tau+1, t+1)])

def cori_denominator(t, v):
    """
    Compute Cori's method denominator
    """    
    # First index s-k = 0: s-k=0; t-tau+1-k=0; t = tau + len(Constants.lambs) - 1
    return 1/Constants.b +  sum([v[s-k]*Constants.lambs[k] for s in range(t-Constants.tau+1, t+1) for k in range(len(Constants.lambs))])
    
def get_Rn_Cori(t, i):
    '''
    Compute R_t with Cori's method
    '''
    l = len(Constants.lambs)
    #assert n >= l
    
    numerator = cori_numerator(t, i)
    denominator = cori_denominator(t, i)

    if denominator < 1e-9: # Avoid numerical problems
        return 0
    
    # Avoid values close to zero
    R = numerator / (denominator +  5/(1+denominator))
    R = min(R, 3.5)
    return R

def get_Rn_curve_Cori(i):
    '''
    Compute the R_t curve for all valid n
    '''
    start, stop = Constants.tau + len(Constants.lambs) - 1, len(i)
    Rn = [get_Rn_Cori(t, i) for t in range(start, stop)]
    return array(Rn)

def get_conf_interval_Cori(i):
    '''
    Obtain a error for the confidence interval of Cori's method
    '''
    start, stop = Constants.tau + len(Constants.lambs) - 1, len(i)

    errors = []
    for k in range(Constants.tau + len(Constants.lambs) - 1):
        errors.append(0)

    for t in range(start, stop):
        numerator = cori_numerator(t, i)
        denominator = cori_denominator(t, i)

        errors.append(sqrt(numerator) / denominator)
    return array(errors)

def create_file_if_doesnt_exist(filename):
    '''
    Create an empty file if the given filename doesn't exist
    '''
    if not isfile(filename):
        open(filename, "wt")


# Parse arguments
parser = argparse.ArgumentParser()
parser.add_argument("country", help="Country name")
parser.add_argument("lambdas_calculation", type=str, help="Lambdas calculation")
parser.add_argument("regularization_weight", type=float, help="initial weight of the regularization term")
parser.add_argument("renewal_equation", type=int, help="Renewal equation", default=1)
parser.add_argument("remove_wa", type=str, help="Remove weekly administrative", default="False")
parser.add_argument("NWeeks", type=int, help="NWeeks", default=8)
parser.add_argument("forecast", type=int, help="Forecast computation")
parser.add_argument("HermiteR", type=float, help="Expected new Rt for forecasting")
parser.add_argument("HermiteNdays", type=int, help="Number of days to reach the new Rt value")
parser.add_argument("--demoExtras", help="DemoExtras directory", default="gi_extras/")
args = parser.parse_args()

is_upload_cases = isfile('input_0.txt')
is_upload_serial_interval = isfile('input_1.txt')

# Copy arguments
country = args.country
data_directory = join(args.demoExtras, "storage")

# Set script constants
ScriptConstants.NWeeks = args.NWeeks

# Update countries in foreground if local storage is empty
if not isfile(join(data_directory, 'cases_FRA.txt')):
    update_in_background(data_directory, background=False)

# Obtain the cases' filename as input_0.txt regardless of its origin
if is_upload_cases:
    # input_0.txt is already there
    copyfile("input_0.txt", "input_cases.txt")
    dates = None
    country = None
else:
    is_aggregated = country.startswith("agg_")
    if is_aggregated:
        if country == "agg_EU":
            agg_countries = ("AUT", "BEL", "BGR", "HRV", "CYP", "CZE", "DEU", "DNK", "EST", "FIN", "FRA", "GRC", "HUN", "IRL", "ITA", "LVA", "LTU", "LUX", "MLT", "NLD", "POL", "PRT", "ROU", "SRB", "SRB", "SVN", "SWE", "ESP")
        elif country == "agg_south_east_asia":
            agg_countries = ("SGP", "MYS", "IDN", "BRN", "THA", "VNM")
        else:
            demo_failure(f"Unknown aggregated '{country}'")

    # Extract counts from daily data
    # d: deaths
    # i: infections
    # dates: dates
    # P: population
    
    if is_aggregated:
        d, i, agg_cases_country, P, dates = aggregate_countries(agg_countries)
        save_country(i, d, min(dates), max(dates), P, country, '.')    
    else:
        d, i, P, dates = read_country(country, data_directory)

    # Save file as input_0.txt regardless of its origin
    filename_from = join('.' if is_aggregated else data_directory, f"cases_{country}.txt")
    copyfile(filename_from, "input_cases.txt")
    
    if is_aggregated:
        # Special case: the C++ wants the multi-column format with all countries
        # Just overwrite it as input_cases.txt
        generate_multicolumn_aggregated("input_cases.txt", agg_cases_country, agg_countries)

# Obtain the serial interval's filename as input_1.txt regardless of its origin
if is_upload_serial_interval:
    copyfile("input_1.txt", "input_lambdas.txt")
else:
    # Not uploaded: copy it from the demoExtras 
    filename_from = join(args.demoExtras, "serial_intervals", args.lambdas_calculation + ".txt")
    copyfile(filename_from, "input_lambdas.txt")

# Call C++ program
remove_wa = 1 if args.remove_wa.lower() == 'true' else 0
program_path = "./coronavirus" if isfile("./coronavirus") else "coronavirus"
command = f"{program_path} input_cases.txt input_lambdas.txt {args.regularization_weight} {args.renewal_equation} {remove_wa} {args.NWeeks} {args.forecast} {args.HermiteR} {args.HermiteNdays}"
print(f"Running: {command}")
code = os.system(command)
assert code == 0, "C++ program runtime error"

# If the C++ program generated a demo_failure.txt, just exit
if isfile("demo_failure.txt"):
    sys.exit(0)

# Read program output and recompute dates
Rn, i, i_f, variabilities, weekend_noise_free = read_program_output("Rt.csv")

# Obtain s_Cori from the text file
s_Cori = read_float("ShiftEpiEstim.txt")
s_Cori = int(np.round(s_Cori))

# We assume that the output signals (Rn, i, i_f, ...) have a variable size.
# Thus, we obtain a list of dates of that size
max_days = get_max_days(Rn, i, i_f)

if dates:
    dates = date_range_num(min(dates), max_days)
    # The last last_data_date starts at the first data and it has the
    # same days as the number of infections.
    #
    # The values in "i" are nan when doing a forecast.
    # Thus, we add number of non-nan values to the first date to know last_data_date
    last_data_date = min(dates) + datetime.timedelta(days=int(sum(~np.isnan(i)))-1)
    save_infections_dates_plot("infections.png", i, i_f, weekend_noise_free, dates, last_data_date, country)
    save_infections_dates_plot("infections.pdf", i, i_f, weekend_noise_free, dates, last_data_date, country)
else:
    last_data_index = int(sum(~np.isnan(i)))-1
    save_infections_days_plot("infections.png", i, i_f, weekend_noise_free, last_data_index, country)
    save_infections_days_plot("infections.pdf", i, i_f, weekend_noise_free, last_data_index, country)
    # If no dates (file upload), then assume last day is today
    #last_data_date = datetime.date.today()
    #start_date = last_data_date - datetime.timedelta(days=max_days-1)
    #dates = date_range(start_date, last_data_date)

# Execute Cori's method

# Read lambdas and normalize
Constants.lambs = read_lambdas_acc("input_lambdas.txt")
Constants.lambs = Constants.lambs / sum(Constants.lambs)

assert Constants.tau % 2 == 1

Rn_Cori = get_Rn_curve_Cori(i)
variability_Cori = get_conf_interval_Cori(Rn_Cori)

# And save plot
if dates:
    save_R_dates_plot(Rn, variabilities, Rn_Cori, variability_Cori, s_Cori, dates, last_data_date, country)
else:
    save_R_days_plot(Rn, variabilities, Rn_Cori, variability_Cori, s_Cori, last_data_index, country)

# If SummaryExecutionOutcomes.txt doesn't exist, create it to avoid that the
# webpage requests a non-existing resource.
create_file_if_doesnt_exist("SummaryExecutionOutcomes.txt")

# Save variables to algo_info.txt
with open("algo_info.txt", "wt") as f:
    f.write(f"s_Cori={s_Cori}\n")

# Update countries in background
update_in_background(data_directory)
